@@ -8,28 +8,27 @@
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2017 by Bradford W. Mott, Stephen Anthony
+// Copyright (c) 1995-2018 by Bradford W. Mott, Stephen Anthony
 // and the Stella Team
 //
 // See the file "License.txt" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
 //============================================================================
 
-#include "bspf.hxx"
-
-#include <iostream>
-#include <fstream>
-#include <sstream>
 #include <map>
 
+#include "bspf.hxx"
+
 #include "Version.hxx"
 #include "OSystem.hxx"
 #include "FrameBuffer.hxx"
+#include "EventHandler.hxx"
 #include "FSNode.hxx"
 #include "Settings.hxx"
 #include "DebuggerDialog.hxx"
 #include "DebuggerParser.hxx"
 #include "StateManager.hxx"
+#include "RewindManager.hxx"
 
 #include "Console.hxx"
 #include "System.hxx"
@@ -58,55 +57,6 @@
 
 Debugger* Debugger::myStaticDebugger = nullptr;
 
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-static const char* const builtin_functions[][3] = {
-  // { "name", "definition", "help text" }
-
-  // left joystick:
-  { "_joy0left", "!(*SWCHA & $40)", "Left joystick moved left" },
-  { "_joy0right", "!(*SWCHA & $80)", "Left joystick moved right" },
-  { "_joy0up", "!(*SWCHA & $10)", "Left joystick moved up" },
-  { "_joy0down", "!(*SWCHA & $20)", "Left joystick moved down" },
-  { "_joy0button", "!(*INPT4 & $80)", "Left joystick button pressed" },
-
-  // right joystick:
-  { "_joy1left", "!(*SWCHA & $04)", "Right joystick moved left" },
-  { "_joy1right", "!(*SWCHA & $08)", "Right joystick moved right" },
-  { "_joy1up", "!(*SWCHA & $01)", "Right joystick moved up" },
-  { "_joy1down", "!(*SWCHA & $02)", "Right joystick moved down" },
-  { "_joy1button", "!(*INPT5 & $80)", "Right joystick button pressed" },
-
-  // console switches:
-  { "_select", "!(*SWCHB & $02)", "Game Select pressed" },
-  { "_reset", "!(*SWCHB & $01)", "Game Reset pressed" },
-  { "_color", "*SWCHB & $08", "Color/BW set to Color" },
-  { "_bw", "!(*SWCHB & $08)", "Color/BW set to BW" },
-  { "_diff0b", "!(*SWCHB & $40)", "Left diff. set to B (easy)" },
-  { "_diff0a", "*SWCHB & $40", "Left diff. set to A (hard)" },
-  { "_diff1b", "!(*SWCHB & $80)", "Right diff. set to B (easy)" },
-  { "_diff1a", "*SWCHB & $80", "Right diff. set to A (hard)" },
-
-  // empty string marks end of list, do not remove
-  { 0, 0, 0 }
-};
-
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-// Names are defined here, but processed in YaccParser
-static const char* const pseudo_registers[][2] = {
-  // { "name", "help text" }
-
-  { "_bank", "Currently selected bank" },
-  { "_rwport", "Address at which a read from a write port occurred" },
-  { "_scan", "Current scanline count" },
-  { "_fcount", "Number of frames since emulation started" },
-  { "_cclocks", "Color clocks on current scanline" },
-  { "_vsync", "Whether vertical sync is enabled (1 or 0)" },
-  { "_vblank", "Whether vertical blank is enabled (1 or 0)" },
-
-  // empty string marks end of list, do not remove
-  { 0, 0 }
-};
-
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Debugger::Debugger(OSystem& osystem, Console& console)
   : DialogContainer(osystem),
@@ -132,6 +82,11 @@ Debugger::Debugger(OSystem& osystem, Console& console)
   myStaticDebugger = this;
 }
 
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Debugger::~Debugger()
+{
+}
+
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void Debugger::initialize()
 {
@@ -152,8 +107,9 @@ void Debugger::initialize()
   myDialog = new DebuggerDialog(myOSystem, *this, 0, 0, myWidth, myHeight);
   myBaseDialog = myDialog;
 
-  myRewindManager = make_unique<RewindManager>(myOSystem, myDialog->rewindButton());
   myCartDebug->setDebugWidget(&(myDialog->cartDebug()));
+
+  saveOldState();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -164,7 +120,7 @@ FBInitStatus Debugger::initializeVideo()
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-bool Debugger::start(const string& message, int address)
+bool Debugger::start(const string& message, int address, bool read)
 {
   if(myOSystem.eventHandler().enterDebugMode())
   {
@@ -173,8 +129,7 @@ bool Debugger::start(const string& message, int address)
     ostringstream buf;
     buf << message;
     if(address > -1)
-      buf << Common::Base::HEX4 << address;
-
+      buf << cartDebug().getLabel(address, read, 4);
     myDialog->message().setText(buf.str());
     return true;
   }
@@ -204,26 +159,26 @@ void Debugger::quit(bool exitrom)
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-string Debugger::autoExec()
+string Debugger::autoExec(StringList* history)
 {
   ostringstream buf;
 
-  // autoexec.stella is always run
-  FilesystemNode autoexec(myOSystem.baseDir() + "autoexec.stella");
+  // autoexec.script is always run
+  FilesystemNode autoexec(myOSystem.baseDir() + "autoexec.script");
   buf << "autoExec():" << endl
-      << myParser->exec(autoexec) << endl;
+      << myParser->exec(autoexec, history) << endl;
 
-  // Also, "romname.stella" if present
-  FilesystemNode romname(myOSystem.romFile().getPathWithExt(".stella"));
-  buf << myParser->exec(romname) << endl;
+  // Also, "romname.script" if present
+  FilesystemNode romname(myOSystem.romFile().getPathWithExt(".script"));
+  buf << myParser->exec(romname, history) << endl;
 
   // Init builtins
-  for(int i = 0; builtin_functions[i][0] != 0; i++)
+  for(uInt32 i = 0; i < NUM_BUILTIN_FUNCS; ++i)
   {
     // TODO - check this for memory leaks
-    int res = YaccParser::parse(builtin_functions[i][1]);
+    int res = YaccParser::parse(ourBuiltinFunctions[i].defn);
     if(res == 0)
-      addFunction(builtin_functions[i][0], builtin_functions[i][1],
+      addFunction(ourBuiltinFunctions[i].name, ourBuiltinFunctions[i].defn,
                   YaccParser::getResult(), true);
     else
       cerr << "ERROR in builtin function!" << endl;
@@ -231,6 +186,24 @@ string Debugger::autoExec()
   return buf.str();
 }
 
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+PackedBitArray& Debugger::breakPoints() const
+{
+  return mySystem.m6502().breakPoints();
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+TrapArray& Debugger::readTraps() const
+{
+  return mySystem.m6502().readTraps();
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+TrapArray& Debugger::writeTraps() const
+{
+  return mySystem.m6502().writeTraps();
+}
+
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 const string Debugger::run(const string& command)
 {
@@ -301,19 +274,18 @@ void Debugger::loadState(int state)
 int Debugger::step()
 {
   saveOldState();
-  mySystem.clearDirtyPages();
 
-  int cyc = mySystem.cycles();
+  uInt64 startCycle = mySystem.cycles();
 
   unlockBankswitchState();
   myOSystem.console().tia().updateScanlineByStep().flushLineCache();
   lockBankswitchState();
 
-  return mySystem.cycles() - cyc;
+  addState("step");
+  return int(mySystem.cycles() - startCycle);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-
 // trace is just like step, except it treats a subroutine call as one
 // instruction.
 
@@ -329,16 +301,16 @@ int Debugger::trace()
   if(mySystem.peek(myCpuDebug->pc()) == 32)
   {
     saveOldState();
-    mySystem.clearDirtyPages();
 
-    int cyc = mySystem.cycles();
+    uInt64 startCycle = mySystem.cycles();
     int targetPC = myCpuDebug->pc() + 3; // return address
 
     unlockBankswitchState();
     myOSystem.console().tia().updateScanlineByTrace(targetPC).flushLineCache();
     lockBankswitchState();
 
-    return mySystem.cycles() - cyc;
+    addState("trace");
+    return int(mySystem.cycles() - startCycle);
   }
   else
     return step();
@@ -366,24 +338,43 @@ bool Debugger::breakPoint(uInt16 bp)
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void Debugger::toggleReadTrap(uInt16 t)
+void Debugger::addReadTrap(uInt16 t)
 {
   readTraps().initialize();
-  readTraps().toggle(t);
+  readTraps().add(t);
 }
 
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void Debugger::toggleWriteTrap(uInt16 t)
+void Debugger::addWriteTrap(uInt16 t)
 {
   writeTraps().initialize();
-  writeTraps().toggle(t);
+  writeTraps().add(t);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void Debugger::toggleTrap(uInt16 t)
+void Debugger::addTrap(uInt16 t)
 {
-  toggleReadTrap(t);
-  toggleWriteTrap(t);
+  addReadTrap(t);
+  addWriteTrap(t);
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+void Debugger::removeReadTrap(uInt16 t)
+{
+  readTraps().initialize();
+  readTraps().remove(t);
+}
+
+void Debugger::removeWriteTrap(uInt16 t)
+{
+  writeTraps().initialize();
+  writeTraps().remove(t);
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+void Debugger::removeTrap(uInt16 t)
+{
+  removeReadTrap(t);
+  removeWriteTrap(t);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -398,11 +389,106 @@ bool Debugger::writeTrap(uInt16 t)
   return writeTraps().isInitialized() && writeTraps().isSet(t);
 }
 
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+uInt8 Debugger::peek(uInt16 addr, uInt8 flags)
+{
+  return mySystem.peek(addr, flags);
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+uInt16 Debugger::dpeek(uInt16 addr, uInt8 flags)
+{
+  return uInt16(mySystem.peek(addr, flags) | (mySystem.peek(addr+1, flags) << 8));
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+void Debugger::poke(uInt16 addr, uInt8 value, uInt8 flags)
+{
+  mySystem.poke(addr, value, flags);
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+M6502& Debugger::m6502() const
+{
+  return mySystem.m6502();
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+int Debugger::peekAsInt(int addr, uInt8 flags)
+{
+  return mySystem.peek(uInt16(addr), flags);
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+int Debugger::dpeekAsInt(int addr, uInt8 flags)
+{
+  return mySystem.peek(uInt16(addr), flags) |
+      (mySystem.peek(uInt16(addr+1), flags) << 8);
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+int Debugger::getAccessFlags(uInt16 addr) const
+{
+  return mySystem.getAccessFlags(addr);
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+void Debugger::setAccessFlags(uInt16 addr, uInt8 flags)
+{
+  mySystem.setAccessFlags(addr, flags);
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+uInt32 Debugger::getBaseAddress(uInt32 addr, bool read)
+{
+  if((addr & 0x1080) == 0x0000) // (addr & 0b 0001 0000 1000 0000) == 0b 0000 0000 0000 0000
+  {
+    if(read)
+      // ADDR_TIA read (%xxx0 xxxx 0xxx ????)
+      return addr & 0x000f; // 0b 0000 0000 0000 1111
+    else
+      // ADDR_TIA write (%xxx0 xxxx 0x?? ????)
+      return addr & 0x003f; // 0b 0000 0000 0011 1111
+  }
+
+  // ADDR_ZPRAM (%xxx0 xx0x 1??? ????)
+  if((addr & 0x1280) == 0x0080) // (addr & 0b 0001 0010 1000 0000) == 0b 0000 0000 1000 0000
+    return addr & 0x00ff; // 0b 0000 0000 1111 1111
+
+  // ADDR_ROM
+  if(addr & 0x1000)
+    return addr & 0x1fff; // 0b 0001 1111 1111 1111
+
+  // ADDR_IO read/write I/O registers (%xxx0 xx1x 1xxx x0??)
+  if((addr & 0x1284) == 0x0280) // (addr & 0b 0001 0010 1000 0100) == 0b 0000 0010 1000 0000
+    return addr & 0x0283; // 0b 0000 0010 1000 0011
+
+  // ADDR_IO write timers (%xxx0 xx1x 1xx1 ?1??)
+  if(!read && (addr & 0x1294) == 0x0294) // (addr & 0b 0001 0010 1001 0100) == 0b 0000 0010 1001 0100
+    return addr & 0x029f; // 0b 0000 0010 1001 1111
+
+  // ADDR_IO read timers (%xxx0 xx1x 1xxx ?1x0)
+  if(read && (addr & 0x1285) == 0x0284) // (addr & 0b 0001 0010 1000 0101) == 0b 0000 0010 1000 0100
+    return addr & 0x028c; // 0b 0000 0010 1000 1100
+
+  // ADDR_IO read timer/PA7 interrupt (%xxx0 xx1x 1xxx x1x1)
+  if(read && (addr & 0x1285) == 0x0285) // (addr & 0b 0001 0010 1000 0101) == 0b 0000 0010 1000 0101
+    return addr & 0x0285; // 0b 0000 0010 1000 0101
+
+  // ADDR_IO write PA7 edge control (%xxx0 xx1x 1xx0 x1??)
+  if(!read && (addr & 0x1294) == 0x0284) // (addr & 0b 0001 0010 1001 0100) == 0b 0000 0010 1000 0100
+    return addr & 0x0287; // 0b 0000 0010 1000 0111
+
+  return 0;
+}
+
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void Debugger::nextScanline(int lines)
 {
+  ostringstream buf;
+  buf << "scanline + " << lines;
+
   saveOldState();
-  mySystem.clearDirtyPages();
 
   unlockBankswitchState();
   while(lines)
@@ -412,14 +498,17 @@ void Debugger::nextScanline(int lines)
   }
   lockBankswitchState();
 
+  addState(buf.str());
   myOSystem.console().tia().flushLineCache();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void Debugger::nextFrame(int frames)
 {
+  ostringstream buf;
+  buf << "frame + " << frames;
+
   saveOldState();
-  mySystem.clearDirtyPages();
 
   unlockBankswitchState();
   while(frames)
@@ -428,18 +517,46 @@ void Debugger::nextFrame(int frames)
     --frames;
   }
   lockBankswitchState();
+
+  addState(buf.str());
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-bool Debugger::rewindState()
+void Debugger::updateRewindbuttons(const RewindManager& r)
 {
-  mySystem.clearDirtyPages();
+  myDialog->rewindButton().setEnabled(!r.atFirst());
+  myDialog->unwindButton().setEnabled(!r.atLast());
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+uInt16 Debugger::windStates(uInt16 numStates, bool unwind, string& message)
+{
+  RewindManager& r = myOSystem.state().rewindManager();
+
+  saveOldState();
 
   unlockBankswitchState();
-  bool result = myRewindManager->rewindState();
+
+  uInt64 startCycles = myOSystem.console().tia().cycles();
+  uInt16 winds = r.windStates(numStates, unwind);
+  message = r.getUnitString(myOSystem.console().tia().cycles() - startCycles);
+
   lockBankswitchState();
 
-  return result;
+  updateRewindbuttons(r);
+  return winds;
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+uInt16 Debugger::rewindStates(const uInt16 numStates, string& message)
+{
+  return windStates(numStates, false, message);
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+uInt16 Debugger::unwindStates(const uInt16 numStates, string& message)
+{
+  return windStates(numStates, true, message);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -461,6 +578,12 @@ string Debugger::showWatches()
   return myParser->showWatches();
 }
 
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+int Debugger::stringToValue(const string& stringval)
+{
+  return myParser->decipher_arg(stringval);
+}
+
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 bool Debugger::patchROM(uInt16 addr, uInt8 value)
 {
@@ -468,15 +591,24 @@ bool Debugger::patchROM(uInt16 addr, uInt8 value)
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void Debugger::saveOldState(bool addrewind)
+void Debugger::saveOldState(bool clearDirtyPages)
 {
+  if (clearDirtyPages)
+    mySystem.clearDirtyPages();
+
   myCartDebug->saveOldState();
   myCpuDebug->saveOldState();
   myRiotDebug->saveOldState();
   myTiaDebug->saveOldState();
+}
 
-  // Add another rewind level to the Undo list
-  if(addrewind)  myRewindManager->addState();
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+void Debugger::addState(string rewindMsg)
+{
+  // Add another rewind level to the Time Machine buffer
+  RewindManager& r = myOSystem.state().rewindManager();
+  r.addState(rewindMsg);
+  updateRewindbuttons(r);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -485,11 +617,14 @@ void Debugger::setStartState()
   // Lock the bus each time the debugger is entered, so we don't disturb anything
   lockBankswitchState();
 
-  // Start a new rewind list
-  myRewindManager->clear();
-
-  // Save initial state, but don't add it to the rewind list
-  saveOldState(false);
+  // Save initial state and add it to the rewind list (except when in currently rewinding)
+  RewindManager& r = myOSystem.state().rewindManager();
+  // avoid invalidating future states when entering the debugger e.g. during rewind
+  if(r.atLast() && (myOSystem.eventHandler().state() != EventHandlerState::TIMEMACHINE
+     || myOSystem.state().mode() == StateManager::Mode::Off))
+    addState("enter debugger");
+  else
+    updateRewindbuttons(r);
 
   // Set the 're-disassemble' flag, but don't do it until the next scheduled time
   myDialog->rom().invalidate(false);
@@ -498,6 +633,8 @@ void Debugger::setStartState()
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void Debugger::setQuitState()
 {
+  saveOldState();
+
   // Bus must be unlocked for normal operation when leaving debugger mode
   unlockBankswitchState();
 
@@ -518,6 +655,15 @@ bool Debugger::addFunction(const string& name, const string& definition,
   return true;
 }
 
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+bool Debugger::isBuiltinFunction(const string& name)
+{
+  for(uInt32 i = 0; i < NUM_BUILTIN_FUNCS; ++i)
+    if(name == ourBuiltinFunctions[i].name)
+      return true;
+  return false;
+}
+
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 bool Debugger::delFunction(const string& name)
 {
@@ -526,8 +672,7 @@ bool Debugger::delFunction(const string& name)
     return false;
 
   // We never want to delete built-in functions
-  for(int i = 0; builtin_functions[i][0] != 0; ++i)
-    if(name == builtin_functions[i][0])
+  if(isBuiltinFunction(name))
       return false;
 
   myFunctions.erase(name);
@@ -567,39 +712,39 @@ string Debugger::builtinHelp() const
   uInt16 len, c_maxlen = 0, i_maxlen = 0;
 
   // Get column widths for aligned output (functions)
-  for(int i = 0; builtin_functions[i][0] != 0; ++i)
+  for(uInt32 i = 0; i < NUM_BUILTIN_FUNCS; ++i)
   {
-    len = strlen(builtin_functions[i][0]);
+    len = ourBuiltinFunctions[i].name.size();
     if(len > c_maxlen)  c_maxlen = len;
-    len = strlen(builtin_functions[i][1]);
+    len = ourBuiltinFunctions[i].defn.size();
     if(len > i_maxlen)  i_maxlen = len;
   }
 
   buf << std::setfill(' ') << endl << "Built-in functions:" << endl;
-  for(int i = 0; builtin_functions[i][0] != 0; ++i)
+  for(uInt32 i = 0; i < NUM_BUILTIN_FUNCS; ++i)
   {
-    buf << std::setw(c_maxlen) << std::left << builtin_functions[i][0]
+    buf << std::setw(c_maxlen) << std::left << ourBuiltinFunctions[i].name
         << std::setw(2) << std::right << "{"
-        << std::setw(i_maxlen) << std::left << builtin_functions[i][1]
+        << std::setw(i_maxlen) << std::left << ourBuiltinFunctions[i].defn
         << std::setw(4) << "}"
-        << builtin_functions[i][2]
+        << ourBuiltinFunctions[i].help
         << endl;
   }
 
   // Get column widths for aligned output (pseudo-registers)
   c_maxlen = 0;
-  for(int i = 0; pseudo_registers[i][0] != 0; ++i)
+  for(uInt32 i = 0; i < NUM_PSEUDO_REGS; ++i)
   {
-    len = strlen(pseudo_registers[i][0]);
+    len = ourPseudoRegisters[i].name.size();
     if(len > c_maxlen)  c_maxlen = len;
   }
 
   buf << endl << "Pseudo-registers:" << endl;
-  for(int i = 0; pseudo_registers[i][0] != 0; ++i)
+  for(uInt32 i = 0; i < NUM_PSEUDO_REGS; ++i)
   {
-    buf << std::setw(c_maxlen) << std::left << pseudo_registers[i][0]
+    buf << std::setw(c_maxlen) << std::left << ourPseudoRegisters[i].name
         << std::setw(2) << " "
-        << std::setw(i_maxlen) << std::left << pseudo_registers[i][1]
+        << std::setw(i_maxlen) << std::left << ourPseudoRegisters[i].help
         << endl;
   }
 
@@ -609,16 +754,20 @@ string Debugger::builtinHelp() const
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void Debugger::getCompletions(const char* in, StringList& list) const
 {
-  for(const auto& iter: myFunctions)
+  // skip if filter equals "_" only
+  if(!BSPF::equalsIgnoreCase(in, "_"))
   {
-    const char* l = iter.first.c_str();
-    if(BSPF::startsWithIgnoreCase(l, in))
-      list.push_back(l);
-  }
+    for(const auto& iter : myFunctions)
+    {
+      const char* l = iter.first.c_str();
+      if(BSPF::matches(l, in))
+        list.push_back(l);
+    }
 
-  for(int i = 0; pseudo_registers[i][0] != 0; ++i)
-    if(BSPF::startsWithIgnoreCase(pseudo_registers[i][0], in))
-      list.push_back(pseudo_registers[i][0]);
+    for(uInt32 i = 0; i < NUM_PSEUDO_REGS; ++i)
+      if(BSPF::matches(ourPseudoRegisters[i].name, in))
+        list.push_back(ourPseudoRegisters[i].name);
+  }
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -636,88 +785,48 @@ void Debugger::unlockBankswitchState()
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Debugger::RewindManager::RewindManager(OSystem& system, ButtonWidget& button)
-  : myOSystem(system),
-    myRewindButton(button),
-    mySize(0),
-    myTop(0)
-{
-  for(int i = 0; i < MAX_SIZE; ++i)
-    myStateList[i] = nullptr;
-}
+Debugger::BuiltinFunction Debugger::ourBuiltinFunctions[NUM_BUILTIN_FUNCS] = {
+  // left joystick:
+  { "_joy0left",    "!(*SWCHA & $40)", "Left joystick moved left" },
+  { "_joy0right",   "!(*SWCHA & $80)", "Left joystick moved right" },
+  { "_joy0up",      "!(*SWCHA & $10)", "Left joystick moved up" },
+  { "_joy0down",    "!(*SWCHA & $20)", "Left joystick moved down" },
+  { "_joy0button",  "!(*INPT4 & $80)", "Left joystick button pressed" },
+
+  // right joystick:
+  { "_joy1left",    "!(*SWCHA & $04)", "Right joystick moved left" },
+  { "_joy1right",   "!(*SWCHA & $08)", "Right joystick moved right" },
+  { "_joy1up",      "!(*SWCHA & $01)", "Right joystick moved up" },
+  { "_joy1down",    "!(*SWCHA & $02)", "Right joystick moved down" },
+  { "_joy1button",  "!(*INPT5 & $80)", "Right joystick button pressed" },
+
+  // console switches:
+  { "_select",    "!(*SWCHB & $02)",  "Game Select pressed" },
+  { "_reset",     "!(*SWCHB & $01)",  "Game Reset pressed" },
+  { "_color",     "*SWCHB & $08",     "Color/BW set to Color" },
+  { "_bw",        "!(*SWCHB & $08)",  "Color/BW set to BW" },
+  { "_diff0b",    "!(*SWCHB & $40)",  "Left diff. set to B (easy)" },
+  { "_diff0a",    "*SWCHB & $40",     "Left diff. set to A (hard)" },
+  { "_diff1b",    "!(*SWCHB & $80)",  "Right diff. set to B (easy)" },
+  { "_diff1a",    "*SWCHB & $80",     "Right diff. set to A (hard)" }
+};
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Debugger::RewindManager::~RewindManager()
-{
-  for(int i = 0; i < MAX_SIZE; ++i)
-    delete myStateList[i];
-}
-
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-bool Debugger::RewindManager::addState()
-{
-  // Create a new Serializer object if we need one
-  if(myStateList[myTop] == nullptr)
-    myStateList[myTop] = new Serializer();
-  Serializer& s = *(myStateList[myTop]);
-
-  if(s)
-  {
-    s.reset();
-    if(myOSystem.state().saveState(s) && myOSystem.console().tia().saveDisplay(s))
-    {
-      // Are we still within the allowable size, or are we overwriting an item?
-      mySize++; if(mySize > MAX_SIZE) mySize = MAX_SIZE;
-
-      myTop = (myTop + 1) % MAX_SIZE;
-      myRewindButton.setEnabled(true);
-      return true;
-    }
-  }
-  return false;
-}
-
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-bool Debugger::RewindManager::rewindState()
-{
-  if(mySize > 0)
-  {
-    mySize--;
-    myTop = myTop == 0 ? MAX_SIZE - 1 : myTop - 1;
-    Serializer& s = *(myStateList[myTop]);
-
-    s.reset();
-    myOSystem.state().loadState(s);
-    myOSystem.console().tia().loadDisplay(s);
-
-    if(mySize == 0)
-      myRewindButton.setEnabled(false);
-
-    return true;
-  }
-  else
-    return false;
-}
-
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-bool Debugger::RewindManager::empty()
-{
-  return mySize == 0;
-}
-
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void Debugger::RewindManager::clear()
-{
-  for(int i = 0; i < MAX_SIZE; ++i)
-    if(myStateList[i] != nullptr)
-      myStateList[i]->reset();
-
-  myTop = mySize = 0;
-
-  // We use Widget::clearFlags here instead of Widget::setEnabled(),
-  // since the latter implies an immediate draw/update, but this method
-  // might be called before any UI exists
-  // TODO - fix this deficiency in the UI core; we shouldn't have to worry
-  //        about such things at this level
-  myRewindButton.clearFlags(WIDGET_ENABLED);
-}
+// Names are defined here, but processed in YaccParser
+Debugger::PseudoRegister Debugger::ourPseudoRegisters[NUM_PSEUDO_REGS] = {
+  { "_bank",      "Currently selected bank" },
+  { "_cclocks",   "Color clocks on current scanline" },
+  { "_cycleshi",  "Higher 32 bits of number of cycles since emulation started" },
+  { "_cycleslo",  "Lower 32 bits of number of cycles since emulation started" },
+  { "_fcount",    "Number of frames since emulation started" },
+  { "_fcycles",   "Number of cycles since frame started" },
+  { "_icycles",    "Number of cycles of last instruction" },
+  { "_rwport",    "Address at which a read from a write port occurred" },
+  { "_scan",      "Current scanline count" },
+  { "_scycles",   "Number of cycles in current scanline" },
+  { "_vblank",    "Whether vertical blank is enabled (1 or 0)" },
+  { "_vsync",     "Whether vertical sync is enabled (1 or 0)" }
+  // CPU address access functions:
+  /*{ "__lastread", "last CPU read address" },
+  { "__lastwrite", "last CPU write address" },*/
+};
